@@ -4,7 +4,7 @@
  * Copyright (C) 1996 Andrew Tridgell
  * Copyright (C) 1996 Paul Mackerras
  * Copyright (C) 2002 Martin Pool <mbp@samba.org>
- * Copyright (C) 2004-2018 Wayne Davison
+ * Copyright (C) 2004-2020 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,7 +29,7 @@ extern int list_only;
 extern int am_sender;
 extern int inc_recurse;
 extern int do_xfers;
-extern int link_dest;
+extern int alt_dest_type;
 extern int preserve_acls;
 extern int preserve_xattrs;
 extern int protocol_version;
@@ -48,6 +48,8 @@ extern struct file_list *cur_flist;
  * we can avoid the pool of dev+inode data.  For incremental recursion mode,
  * the receiver will use a ndx hash to remember old pathnames. */
 
+static void *data_when_new = "";
+
 static struct hashtable *dev_tbl;
 
 static struct hashtable *prior_hlinks;
@@ -57,32 +59,29 @@ static struct file_list *hlink_flist;
 void init_hard_links(void)
 {
 	if (am_sender || protocol_version < 30)
-		dev_tbl = hashtable_create(16, 1);
+		dev_tbl = hashtable_create(16, HT_KEY64);
 	else if (inc_recurse)
-		prior_hlinks = hashtable_create(1024, 0);
+		prior_hlinks = hashtable_create(1024, HT_KEY32);
 }
 
 struct ht_int64_node *idev_find(int64 dev, int64 ino)
 {
 	static struct ht_int64_node *dev_node = NULL;
-	struct hashtable *tbl;
 
 	/* Note that some OSes have a dev == 0, so increment to avoid storing a 0. */
 	if (!dev_node || dev_node->key != dev+1) {
 		/* We keep a separate hash table of inodes for every device. */
-		dev_node = hashtable_find(dev_tbl, dev+1, 1);
-		if (!(tbl = dev_node->data)) {
-			tbl = dev_node->data = hashtable_create(512, 1);
+		dev_node = hashtable_find(dev_tbl, dev+1, data_when_new);
+		if (dev_node->data == data_when_new) {
+			dev_node->data = hashtable_create(512, HT_KEY64);
 			if (DEBUG_GTE(HLINK, 3)) {
-				rprintf(FINFO,
-				    "[%s] created hashtable for dev %s\n",
-				    who_am_i(), big_num(dev));
+				rprintf(FINFO, "[%s] created hashtable for dev %s\n",
+					who_am_i(), big_num(dev));
 			}
 		}
-	} else
-		tbl = dev_node->data;
+	}
 
-	return hashtable_find(tbl, ino, 1);
+	return hashtable_find(dev_node->data, ino, (void*)-1L);
 }
 
 void idev_destroy(void)
@@ -118,15 +117,14 @@ static void match_gnums(int32 *ndx_list, int ndx_count)
 	struct ht_int32_node *node = NULL;
 	int32 gnum, gnum_next;
 
-	qsort(ndx_list, ndx_count, sizeof ndx_list[0],
-	     (int (*)()) hlink_compare_gnum);
+	qsort(ndx_list, ndx_count, sizeof ndx_list[0], (int (*)()) hlink_compare_gnum);
 
 	for (from = 0; from < ndx_count; from++) {
 		file = hlink_flist->sorted[ndx_list[from]];
 		gnum = F_HL_GNUM(file);
 		if (inc_recurse) {
-			node = hashtable_find(prior_hlinks, gnum, 1);
-			if (!node->data) {
+			node = hashtable_find(prior_hlinks, gnum, data_when_new);
+			if (node->data == data_when_new) {
 				if (!(node->data = new_array0(char, 5)))
 					out_of_memory("match_gnums");
 				assert(gnum >= hlink_flist->ndx_start);
@@ -269,7 +267,7 @@ static char *check_prior(struct file_struct *file, int gnum,
 	}
 
 	if (inc_recurse
-	 && (node = hashtable_find(prior_hlinks, gnum, 0)) != NULL) {
+	 && (node = hashtable_find(prior_hlinks, gnum, NULL)) != NULL) {
 		assert(node->data != NULL);
 		if (CVAL(node->data, 0) != 0) {
 			*prev_ndx_p = -1;
@@ -396,7 +394,7 @@ int hard_link_check(struct file_struct *file, int ndx, char *fname,
 			pathjoin(cmpbuf, MAXPATHLEN, basis_dir[j], fname);
 			if (link_stat(cmpbuf, &alt_sx.st, 0) < 0)
 				continue;
-			if (link_dest) {
+			if (alt_dest_type == LINK_DEST) {
 				if (prev_st.st_dev != alt_sx.st.st_dev
 				 || prev_st.st_ino != alt_sx.st.st_ino)
 					continue;
@@ -528,7 +526,7 @@ void finish_hard_link(struct file_struct *file, const char *fname, int fin_ndx,
 
 	if (inc_recurse) {
 		int gnum = F_HL_GNUM(file);
-		struct ht_int32_node *node = hashtable_find(prior_hlinks, gnum, 0);
+		struct ht_int32_node *node = hashtable_find(prior_hlinks, gnum, NULL);
 		if (node == NULL) {
 			rprintf(FERROR, "Unable to find a hlink node for %d (%s)\n", gnum, f_name(file, prev_name));
 			exit_cleanup(RERR_MESSAGEIO);
