@@ -170,6 +170,7 @@ char *logfile_name = NULL;
 char *logfile_format = NULL;
 char *stdout_format = NULL;
 char *password_file = NULL;
+char *early_input_file = NULL;
 char *rsync_path = RSYNC_PATH;
 char *backup_dir = NULL;
 char backup_dir_buf[MAXPATHLEN];
@@ -180,12 +181,17 @@ int rsync_port = 0;
 int alt_dest_type = 0;
 int basis_dir_cnt = 0;
 
+static int version_opt_cnt = 0;
 static int remote_option_alloc = 0;
 int remote_option_cnt = 0;
 const char **remote_options = NULL;
 const char *checksum_choice = NULL;
 const char *compress_choice = NULL;
 
+#ifndef __APPLE__ /* Do we need a configure check for this? */
+#define SUPPORT_ATIMES 1
+#endif
+
 int quiet = 0;
 int output_motd = 1;
 int log_before_transfer = 0;
@@ -602,6 +608,11 @@ static void print_capabilities(enum logcode f)
 #endif
 			"IPv6",
 
+#ifndef SUPPORT_ATIMES
+		"no "
+#endif
+			"atimes",
+
 		"batchfiles",
 
 #ifndef HAVE_FTRUNCATE
@@ -624,6 +635,13 @@ static void print_capabilities(enum logcode f)
 #endif
 			"xattrs",
 
+#ifdef RSYNC_USE_PROTECTED_ARGS
+		"default "
+#else
+		"optional "
+#endif
+			"protect-args",
+
 #ifndef ICONV_OPTION
 		"no "
 #endif
@@ -639,14 +657,38 @@ static void print_capabilities(enum logcode f)
 #endif
 			"prealloc",
 
+	"*" /* All options after this point are hidden w/o -V -V */
 #ifndef HAVE_SIMD
 		"no "
 #endif
 			"SIMD",
 
+#ifndef HAVE_ASM
+		"no "
+#endif
+			"asm",
+
+#ifndef USE_OPENSSL
+		"no "
+#endif
+			"openssl-crypto",
+
 		NULL
 	};
 
+	for (line_len = 0, j = 0; ; j++) {
+		char *cap = capabilities[j];
+		if (!cap)
+			break;
+		if (*cap == '*') {
+			if (version_opt_cnt >= 2)
+				capabilities[j]++;
+			else
+				capabilities[j] = NULL;
+			break;
+		}
+	}
+
 	for (line_len = 0, j = 0; ; j++) {
 		char *cap = capabilities[j];
 		int cap_len = cap ? strlen(cap) : 1000;
@@ -672,7 +714,7 @@ static void print_rsync_version(enum logcode f)
 		RSYNC_NAME, RSYNC_VERSION, PROTOCOL_VERSION, subprotocol);
 
 	rprintf(f, "Copyright (C) 1996-" LATEST_YEAR " by Andrew Tridgell, Wayne Davison, and others.\n");
-	rprintf(f, "Web site: http://rsync.samba.org/\n");
+	rprintf(f, "Web site: https://rsync.samba.org/\n");
 
 	rprintf(f, "Capabilities:\n");
 	print_capabilities(f);
@@ -729,7 +771,7 @@ void usage(enum logcode F)
   rprintf(F,"\n");
   rprintf(F,"Use \"rsync --daemon --help\" to see the daemon-mode command-line options.\n");
   rprintf(F,"Please see the rsync(1) and rsyncd.conf(5) man pages for full documentation.\n");
-  rprintf(F,"See http://rsync.samba.org/ for updates, bug reports, and answers\n");
+  rprintf(F,"See https://rsync.samba.org/ for updates, bug reports, and answers\n");
 }
 
 enum {OPT_SERVER = 1000, OPT_DAEMON, OPT_SENDER, OPT_EXCLUDE, OPT_EXCLUDE_FROM,
@@ -903,6 +945,7 @@ static struct poptOption long_options[] = {
   {"zc",               0,  POPT_ARG_STRING, &compress_choice, 0, 0, 0 },
   {"skip-compress",    0,  POPT_ARG_STRING, &skip_compress, 0, 0, 0 },
   {"compress-level",   0,  POPT_ARG_INT,    &do_compression_level, 0, 0, 0 },
+  {"zl",               0,  POPT_ARG_INT,    &do_compression_level, 0, 0, 0 },
   {0,                 'P', POPT_ARG_NONE,   0, 'P', 0, 0 },
   {"progress",         0,  POPT_ARG_VAL,    &do_progress, 1, 0, 0 },
   {"no-progress",      0,  POPT_ARG_VAL,    &do_progress, 0, 0, 0 },
@@ -962,6 +1005,7 @@ static struct poptOption long_options[] = {
   {"port",             0,  POPT_ARG_INT,    &rsync_port, 0, 0, 0 },
   {"sockopts",         0,  POPT_ARG_STRING, &sockopts, 0, 0, 0 },
   {"password-file",    0,  POPT_ARG_STRING, &password_file, 0, 0, 0 },
+  {"early-input",      0,  POPT_ARG_STRING, &early_input_file, 0, 0, 0 },
   {"blocking-io",      0,  POPT_ARG_VAL,    &blocking_io, 1, 0, 0 },
   {"no-blocking-io",   0,  POPT_ARG_VAL,    &blocking_io, 0, 0, 0 },
   {"outbuf",           0,  POPT_ARG_STRING, &outbuf_mode, 0, 0, 0 },
@@ -1143,6 +1187,9 @@ static void set_refuse_options(void)
 		parse_one_refuse_match(0, "log-file*", list_end);
 	}
 
+#ifndef SUPPORT_ATIMES
+	parse_one_refuse_match(0, "atimes", list_end);
+#endif
 #ifndef SUPPORT_HARD_LINKS
 	parse_one_refuse_match(0, "link-dest", list_end);
 #endif
@@ -1361,8 +1408,8 @@ int parse_arguments(int *argc_p, const char ***argv_p)
 
 		switch (opt) {
 		case 'V':
-			print_rsync_version(FINFO);
-			exit_cleanup(0);
+			version_opt_cnt++;
+			break;
 
 		case OPT_SERVER:
 			if (!am_server) {
@@ -1848,6 +1895,11 @@ int parse_arguments(int *argc_p, const char ***argv_p)
 		}
 	}
 
+	if (version_opt_cnt) {
+		print_rsync_version(FINFO);
+		exit_cleanup(0);
+	}
+
 	if (protect_args < 0) {
 		if (am_server)
 			protect_args = 0;
@@ -2171,6 +2223,7 @@ int parse_arguments(int *argc_p, const char ***argv_p)
 	}
 	if (backup_dir) {
 		size_t len;
+		make_backups = 1; /* --backup-dir implies --backup */
 		while (*backup_dir == '.' && backup_dir[1] == '/')
 			backup_dir += 2;
 		if (*backup_dir == '.' && backup_dir[1] == '\0')
